/*!
 * SlimWidgets JavaScript Library v1.0
 *
 * Dependencies: jQuery
 *
 * A small JS-library containing the following UI-widgets:
 *
 * BackToTopBtn
 * StickyHeader
 * Overlay
 * LoadingSpinner
 * Popup
 * Datepicker
 * Tooltip
 * StaticTooltip
 *
 */
 
(function ($) {

    'use strict';

    /* --------------------------------------------------------------------------------------------------------------
     * Container object + counter
     * -------------------------------------------------------------------------------------------------------------- */

    var SlimWidgets = {}
    
    // Count all created SlimWidgets
    SlimWidgets.Counter = {
    
        backToTopBtn: 0, 
        stickyHeader: 0, 
        overlay: 0,
        loadingSpinner: 0,
        popup: 0,
        ajaxPopup: 0,
        datepicker: 0,
        copyTextPopup: 0,
        tooltip: 0,
        staticTooltip: 0
    }
    
    /* --------------------------------------------------------------------------------------------------------------
     * Options
     * -------------------------------------------------------------------------------------------------------------- */

    /* Protected options
     * -------------------------------------------------------------------------------------------------------------- */

    // Protected options. Not expected to be overridden. 
    // Clasnames defined here might serve as selectors later on.
    SlimWidgets.ProtectedOptions = {
        
        backToTopBtn: {
            _tagName: 'span',
            _className: 'slim-back-to-top-btn'
        }, 
        
        stickyHeader: {
            _className: 'slim-sticky-header'
        },
        
        overlay: {
            _tagName: 'div',
            _className: 'slim-overlay'
        },
        
        loadingSpinner: {
            _tagName: 'div',
            _className: 'slim-loading-spinner-overlay'
        },
        
        popup: {
            _tagName: 'div',
            _className: 'slim-popup-overlay',
            container: {
                _tagName: 'div',
                _className: 'slim-popup-container',
                minMarginY: 40,
                closeBtn: {
                    _tagName: 'span',
                    _className: 'slim-close-btn'  
                },
                title: {
                    _tagName: 'div',
                    _className: 'slim-popup-title-wrapper'    
                },
                content: {
                    _tagName: 'div',
                    _className: 'slim-popup-content-wrapper'    
                }
            }
        },
        
        datepicker: {
            _tagName: 'div',
            _className: 'slim-popup-overlay',
            container: {
                _tagName: 'div',
                _className: 'slim-datepicker-container',
                closeBtn: {
                    _tagName: 'span',
                    _className: 'slim-close-btn'  
                },
                title: {
                    _tagName: 'div',
                    _className: 'slim-datepicker-title-wrapper'    
                },
                dateSelectsWrapper: {
                    _tagName: 'div',
                    _className: 'slim-dateselects-wrapper', 
                    yearSelect: {
                        _tagName: 'select', 
                        _className: 'year-select'
                    },
                    monthSelect: {
                        _tagName: 'select', 
                        _className: 'month-select'
                    },
                    daySelect: {
                        _tagName: 'select', 
                        _className: 'day-select'
                    }
                },
                confirmBtn: {
                    _tagName: 'button', 
                    _className: 'confirm-btn btn btn-fh'
                }, 
                cancelBtn: {
                    _tagName: 'button', 
                    _className: 'cancel-btn btn btn-fh-bright'                    
                }
            }
        },
        
        tooltip: {
            _tagName: 'div',
            _className: 'slim-tooltip',
            _addClassNames: {
                withCloseBtn: 'with-close-btn',
                withArrow: 'with-arrow'
            },
            container: {
                _tagName: 'div',
                _className: 'slim-tooltip-container'
            },
            closeBtn: {
                _tagName: 'span',
                _className: 'slim-close-btn'  
            }
        }
    };
    
    /* Default options
     * -------------------------------------------------------------------------------------------------------------- */

    // Default options that can be overridden by custom user settings when initializing SlimWidgets objects
    SlimWidgets.DefaultOptions = {
    
        backToTopBtn : {
            className: '',
            showAfter: 200,
            scrollSpeed: 300,
            offsetTop: 0
        },
        
        stickyHeader: {
            className: '',
            borderTop: 0,
            borderBottom: Infinity,
            outerWidth: 'inherit',
            left: null
        },
        
        overlay: {
            className: '',
            toDisable: '',          // Element(s) to disable while overlay is shown.
            hideOnClick: false,	    // Remove overlay when clicking it
            hideOnEsc: false       // Remove overlay when pressing escape
        },
        
        loadingSpinner: {
            className: '',
            toDisable: '',          // Element(s) to disable while overlay is shown.
            hideOnClick: false,	    // Remove overlay when clicking it
            hideOnEsc: false,       // Remove overlay when pressing escape
            spinner: {
                className: 'spinner-0'
            }
        },
        
        popup: {
            className: '',
            toDisable: '',          // Element(s) to disable while overlay is shown.
            hideOnClick: true,	    // Remove overlay when clicking it
            hideOnEsc: true,       // Remove overlay when pressing escape
            withCloseBtn: true
        },
        
        datepicker: {
            className: '',
            toDisable: '',          // Element(s) to disable while overlay is shown.
            hideOnClick: true,	    // Remove overlay when clicking it
            hideOnEsc: true,        // Remove overlay when pressing escape
            withCloseBtn: true,
            lang: 'en',
            format: 'YYYY-MM-DD',
            yearStart: (function(){ return (new Date()).getFullYear() -15; }()),
            yearEnd: (function(){ return (new Date()).getFullYear(); }()),
            monthNames: [
                'January',
                'February',
                'March ',
                'April',
                'May',
                'June',
                'July',
                'August',
                'September',
                'October',
                'November',
                'December'
            ],
            textConfirmBtn: 'Go',
            textCancelBtn: 'Cancel',
            invalidDateMsg: 'Invalid date',
            updateOnChange: true
        },
        
        tooltip: {
            className: '',
            offset:{
                top: 5,
                left: 10    
            },
            delay: 800,
            hideOnClick: true,
            withCloseBtn: false,
            withArrow: false
        }
    };
    
    /* Language options
     * -------------------------------------------------------------------------------------------------------------- */

    // Language options
    SlimWidgets.LanguageOptions = {
    
         datepickerGerman: {
            format: 'DD.MM.YYYY', 
            monthNames: [
                'Januar',
                'Februar',
                'März',
                'April',
                'Mai',
                'Juni',
                'Juli',
                'August',
                'September',
                'Oktober',
                'November',
                'Dezember'
            ],
            textConfirmBtn: 'Fertig',
            textCancelBtn: 'Löschen',
            invalidDateMsg: 'Ungültiges Datum'
        }     
    };

    /* --------------------------------------------------------------------------------------------------------------
     * Util
     * -------------------------------------------------------------------------------------------------------------- */
    
    // Utiliy functions
    SlimWidgets.Util = {
    
        // Create a new object with the given prototype
        // See: https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/Object/create)
        createObject: Object.create || function(proto){
            function ctor() {}
            ctor.prototype = proto;
            return new ctor();
        },
        
        // Merge constant options, default options and custom (user) options into a new options object
        buildOptions: function(slimName, customOptions){
            return $.extend(true, {}, SlimWidgets.ProtectedOptions[slimName], SlimWidgets.DefaultOptions[slimName], customOptions);
        }, 
        
        createId: function(slimName){
            SlimWidgets.Counter[slimName] += 1;
            return slimName + SlimWidgets.Counter[slimName];
        },
        
        create$Obj: function(options){
            return $(document.createElement(options._tagName))
                .addClass(options._className);
        },

        createMain$Obj: function(options){
            
            return $(document.createElement(options._tagName))
                .addClass(options._className)
                .addClass(options.className)
                .hide();
        },
        
        unbindWidgetFrom: function(selector, event, slimName){
            
            $(selector).each(function(i, element){
            
                if ($(element).data(slimName)){
                
                    $(element).data(slimName).remove(); // Remove widget
                    $.removeData(element, slimName);    // Remove reference in element
                }
                
                $(element).off(event);  // Remove handler
            });
        }  
    };
    
    
    /* --------------------------------------------------------------------------------------------------------------
     * Core
     * -------------------------------------------------------------------------------------------------------------- */
    
    // Create 
    SlimWidgets.Core = function(slimName, options){
        
        if (!slimName){ return false; }
        
        this.slimName = slimName;    
        this.slimId = SlimWidgets.Util.createId(slimName); 
        this.options = SlimWidgets.Util.buildOptions(slimName, options);
        this.$Objs = {};
        
        if (this.options._tagName){ 
            this.$Objs.main = SlimWidgets.Util.createMain$Obj(this.options);
        } 
    };
    
    
    /* --------------------------------------------------------------------------------------------------------------
     * Back-to-top-button
     * -------------------------------------------------------------------------------------------------------------- */
    
    // Create back-to-top-button
    SlimWidgets.BackToTopBtn = function(options){
        SlimWidgets.Core.call(this, 'backToTopBtn', options);
    };  
       
    // Append   
    SlimWidgets.BackToTopBtn.prototype.append = function(){
        
        var that = this;
        var $backToTopBtn = this.$Objs.main;
        
        // Scroll-to-top handler            
        $backToTopBtn.on('click', function(){
            $('html, body').animate({scrollTop: 0 + that.options.offsetTop}, that.options.scrollSpeed);    
        }).appendTo('body'); 
        
        // Toggle back-to-top-button handler
        $(window).on('scroll.' + this.slimId, function(){
            
            if ($(window).scrollTop() > that.options.showAfter){
                $backToTopBtn.fadeIn();
            } else {
                $backToTopBtn.fadeOut();
            }
            
        }).scroll();  
        
        return this;
    };
    
    // Remove
    SlimWidgets.BackToTopBtn.prototype.remove = function(){
        
        $(window).off('scroll.' + this.slimId); 
        this.$Objs.main.remove();
        
        return this;
    };


    /* --------------------------------------------------------------------------------------------------------------
     * Sticky-Header
     * -------------------------------------------------------------------------------------------------------------- */
    
    // Transforms a custom HTML element into a "sticky" header that sticks to the page top if the page is scrolled.
    SlimWidgets.StickyHeader = function(headerSelector, options){
    
        SlimWidgets.Core.call(this, 'stickyHeader', options);

        this.$Objs.main = $(headerSelector).clone()
            .addClass(this.options._className)
            .addClass(this.options.className)
            .outerWidth(this.options.outerWidth)
            .hide();
        
        if (this.options.left){
            this.$Objs.main.css('left', this.options.left);
        }
    };
    
    // Append sticky header
    SlimWidgets.StickyHeader.prototype.appendTo = function(selector){
        
        var that = this;
        var $stickyHeader = this.$Objs.main; 
        
        $stickyHeader.appendTo(selector); 
        
        // Toggle sticky header handler
        $(window).on('scroll.' + this.slimId, function(){
            
            var currentScroll = $(window).scrollTop();
            
            if (currentScroll > that.options.borderTop && currentScroll < that.options.borderBottom){            
                $stickyHeader.show();       
            } else {
                $stickyHeader.hide();
            }
        }).scroll(); 
        
        return this;
    };
    
    // Update visibility borders and/or outerWidth and/or left position of the sticky header
    SlimWidgets.StickyHeader.prototype.update = function(options){

        if (options.borderTop >= 0){
            this.options.borderTop = options.borderTop;
        } 
        
        if (options.borderBottom){
            this.options.borderBottom = options.borderBottom;
        } 
        
        if (options.outerWidth){
            this.options.outerWidth = options.outerWidth; 
            this.$Objs.main.outerWidth(options.outerWidth); 
        } 
        
        if (options.left){
            this.options.left = options.left; 
            this.$Objs.main.css('left', options.left); 
        } 
        
        return this;
    };
    
    // Adapt visibility borders and outerWidth of the sticky header to a container element
    SlimWidgets.StickyHeader.prototype.adaptToContainer = function(selector, customOffset){
        
        var $container = $(selector); 
        var defaultOffset = {top:0, bottom:0, vertical:0}
        var offset = $.extend({}, defaultOffset, customOffset);
        
        this.update({
            borderTop: $container.offset().top  + offset.top,
            borderBottom: $container.offset().top + $container.innerHeight() + offset.bottom,
            outerWidth: $container.outerWidth() + offset.vertical -2
        });
        
        return this;
    };
    
    // Remove
    SlimWidgets.StickyHeader.prototype.remove = function(){
        
        $(window).off('scroll.' + this.slimId); 
        this.$Objs.main.remove();
        
        return this;
    };


    /* --------------------------------------------------------------------------------------------------------------
     * OVERLAY
     * -------------------------------------------------------------------------------------------------------------- */

    /* An overlay element covering the entire screen. Prevents clicks on underlaying elements. */
    SlimWidgets.Overlay = function(options){
        SlimWidgets.Core.call(this, 'overlay', options);
    };
    
    // Append    
    SlimWidgets.Overlay.prototype.append = function(){

        var that = this; 
        var $overlay = this.$Objs.main;
        
        $overlay.appendTo('body').show(); 
        
        // Optionally disable e.g. a submit button which would otherwise e.g. react to key-press events
        if (this.options.toDisable){
            $(this.options.toDisable).prop('disabled', true).addClass('disabled'); 
        }
        
        // Hide overlay when clicking it
        if (this.options.hideOnClick){
        
            $overlay.on('click', function(e){
                if (e.target === this){ that.remove(); }
            });
        } 
        
        // Hide overlay when pressing escape
        if (this.options.hideOnEsc){
            
            var keyCode = 27;   // 27 = Escape
            
            $(window).on('keydown.' + this.slimId, function(e){
            
                if (e.which == keyCode || e.key == keyCode || e.keyIdentifier == keyCode || e.keyCode == keyCode){  
    
                    // Remove overlay only if it is the currently highest/uppermost one 
                    if ($('.' + that.options._className + ':visible').last().get(0) === that.$Objs.main.get(0)){     
                        that.remove()
                    }
                }
            });
        } 
        
        if (this._hooks){
            this._hooks();
        }
        
        // Prevent page scroll 
        $('body').addClass('no-scroll-' + this.slimId);
        
        return this;
    };    
    
    // Remove
    SlimWidgets.Overlay.prototype.remove = function(s, a){
    
        $(window).off('keydown.' + this.slimId); 
        this.$Objs.main.remove();
        
        // Re-enable page scroll 
        $('html, body').removeClass('no-scroll-'+ this.slimId);
        
        return this;
    };
    
    SlimWidgets.Overlay.prototype.removeOnClickTo = function($target){
        
        if($target && $target.length !== 0){
            
            var that = this; 
            
            $target.one('click', function(e){ 
                e.stopPropagation(); 
                that.remove();
            }); 
        }
    }; 
    
    
    /* --------------------------------------------------------------------------------------------------------------
     * LOADING SPINNER
     * -------------------------------------------------------------------------------------------------------------- */

    /* Extends "Overlay". Adds a single element CSS-loading-spinner to it. All loading spinners are taken from 
     * https://projects.lukehaas.me/css-loaders/.*/ 
    SlimWidgets.LoadingSpinner = function(options){
    
        SlimWidgets.Core.call(this, 'loadingSpinner', options);
        this.$Objs.main.append('<div class="' + this.options.spinner.className + '"/>');
    };

    // Inherit prototype from SlimWidgets.Overlay
    SlimWidgets.LoadingSpinner.prototype = SlimWidgets.Util.createObject(SlimWidgets.Overlay.prototype);
    SlimWidgets.LoadingSpinner.prototype.constructor = SlimWidgets.LoadingSpinner;     
    

    /* --------------------------------------------------------------------------------------------------------------
     * Popup
     * -------------------------------------------------------------------------------------------------------------- */

    /* Extends "Overlay". Adds a popup for custom content. initHooks: [{fn: function(args){...}, args: [arg1, arg2]}, ... ] */
    SlimWidgets.Popup = function(contentElements, options, initHooks){
    
        SlimWidgets.Core.call(this, 'popup', options);
        
        this.$Objs.container = SlimWidgets.Util.create$Obj(this.options.container);
        this.$Objs.main.append(this.$Objs.container);
        
        this.initHooks = initHooks;
        
        this.$Objs.closeBtn;
        this.$Objs.titleWrapper;
        this.$Objs.contentWrapper;
        
        if (this.options.withCloseBtn){
        
            this.$Objs.closeBtn = SlimWidgets.Util.create$Obj(this.options.container.closeBtn);
            this.$Objs.container.append(this.$Objs.closeBtn);
        }
        
        if (contentElements){
        
            this.setTitle(contentElements.title); 
            this.setContent(contentElements.content);
        }
        
        $(window).on('resize.' + this.slimId, $.proxy(this.adjustHeight, this));
    };

    // Inherit prototype from SlimWidgets.Overlay
    SlimWidgets.Popup.prototype = SlimWidgets.Util.createObject(SlimWidgets.Overlay.prototype);
    SlimWidgets.Popup.prototype.constructor = SlimWidgets.Popup;  
    
    // Activate  close button (if any) and optionally invoke custom hooks. All handlers defined here will be removed when function Popup.remove is called 
    // (by jQuery.remove). They will be re-appended when function Popup.append is called (which re-envokes _hooks)
    SlimWidgets.Popup.prototype._hooks = function(){
    
        var that = this;
    
        this.removeOnClickTo(this.$Objs.closeBtn); 
        
        // initHooks: [{fn: function(args){...}, args: [arg1, arg2]}, ... ]
        $.each(that.initHooks, function(i, hook){   
            hook.fn.apply(that, hook.args); 
        });
    };
    
    // Get data about popup container size
    SlimWidgets.Popup.prototype._getContainerData = function(){ 
    
        var data = {}; 
        var $clone = this.$Objs.main.clone();
        var $cloneContainer = $clone.children('.' + this.options.container._className); 
        
        $clone.css('visibility', 'hidden').appendTo('body').show(); 
        data.outerWidth = $cloneContainer.outerWidth();
        data.outerHeight = $cloneContainer.outerHeight();
        data.offset = $cloneContainer.offset();
        $clone.remove(); 
        
        return data;
    };
    
    // Limit popup height to window height and reposition close button if neccessary. Use when window height changes!
    SlimWidgets.Popup.prototype.adjustHeight = function(){  
    
        this.$Objs.container.outerHeight('');       // Reset in case of previous change

        var windowHeight = $(window).height();
        var minMarginY = (this.options.container.minMarginY || 0) * 2;
        var containerData = this._getContainerData(); 
         
        if(containerData.outerHeight + minMarginY > windowHeight){  
            
            // Restrict container height
            this.$Objs.container.outerHeight(windowHeight - minMarginY);
            
            // Modify close button to keep it visible (if any)
            if (this.$Objs.closeBtn){
            
                this.$Objs.closeBtn.css({
                    right: 'auto',
                    left: containerData.offset.left + containerData.outerWidth - 38,
                    top: (this.options.container.minMarginY || 0) + 6
                }).appendTo(this.$Objs.main);
            } 
            
        } else if (this.$Objs.closeBtn){
            
            // Reset close btn if it has been previously modified 
            if (this.$Objs.main.children('.' + this.options.container.closeBtn._className).length !== 0){

                this.$Objs.closeBtn.css({
                    right: '',
                    left: '',
                    top: ''
                }).prependTo(this.$Objs.container);
            }
        } 
        
        return this;
    };
    
    // Set title
    SlimWidgets.Popup.prototype.setTitle = function(title){ 

        if (!title || title.length === 0){ return this; }
        
        // Insert title into existing title wrapper or ...
        if (this.$Objs.titleWrapper){
        
            this.$Objs.titleWrapper.empty().append(title); 
            
        } else {    // ... build new title wrapper and append title
        
            this.$Objs.titleWrapper = SlimWidgets.Util.create$Obj(this.options.container.title).append(title);
            
            if (this.$Objs.closeBtn){
                this.$Objs.titleWrapper.insertAfter(this.$Objs.closeBtn); 
            } else {
                this.$Objs.container.prepend(this.$Objs.titleWrapper);
            } 
        }
        
        this.adjustHeight();

        return this;
    };
    
    // Set content
    SlimWidgets.Popup.prototype.setContent = function(content){ 

        if (!content || content.length === 0){ return this; }
        
        // Insert title into existing title wrapper or ...
        if (this.$Objs.contentWrapper){
        
            this.$Objs.contentWrapper.empty().append(content); 
            
        } else {    // ... build new title wrapper and append title
        
            this.$Objs.contentWrapper = SlimWidgets.Util.create$Obj(this.options.container.content).append(content);
            this.$Objs.container.append(this.$Objs.contentWrapper); 
        }
        
        this.adjustHeight();    

        return this;
    };
    
    // Remove popup content
    SlimWidgets.Popup.prototype.removeContent = function(){ 
    
        if (this.$Objs.contentWrapper){
            this.$Objs.contentWrapper.remove(); 
            this.$Objs.contentWrapper = (function(){}());
        }
        
        $(window).off('resize.' + this.slimId);
    
        return this;
    };
    
    
    /* --------------------------------------------------------------------------------------------------------------
     * Datepicker
     * -------------------------------------------------------------------------------------------------------------- */

    /* Extends "Overlay" to an datepicker popup that allows to insert a selected date into a custom input field. */
    SlimWidgets.Datepicker = function(title, options){
    
        SlimWidgets.Core.call(this, 'datepicker', options);
        
        this.$Objs.container = SlimWidgets.Util.create$Obj(this.options.container);
        this.$Objs.main.append(this.$Objs.container);
        
        this.$Objs.closeBtn;
        this.$Objs.titleWrapper;
        this.$Objs.target;  // Element assigned to/controlled by the datepicker
        
        if (this.options.withCloseBtn){
        
            this.$Objs.closeBtn = SlimWidgets.Util.create$Obj(this.options.container.closeBtn);
            this.$Objs.container.append(this.$Objs.closeBtn);
        }
        
        if (title){
        
            this.$Objs.titleWrapper = SlimWidgets.Util.create$Obj(this.options.container.title).append(title);
            
            if (this.$Objs.closeBtn){
                this.$Objs.titleWrapper.insertAfter(this.$Objs.closeBtn); 
            } else {
                this.$Objs.container.prepend(this.$Objs.titleWrapper);
            } 
        }
        
        this.$Objs.dateSelectsWrapper = SlimWidgets.Util.create$Obj(this.options.container.dateSelectsWrapper);
        this.$Objs.container.append(this.$Objs.dateSelectsWrapper); 
        
        // Build year select 
        this.$Objs.yearSelect = SlimWidgets.Util.create$Obj(this.options.container.dateSelectsWrapper.yearSelect);
         
        for (var y = this.options.yearStart; y <= this.options.yearEnd; y++){
            this.$Objs.yearSelect.append('<option value="' + y + '">' + y + '</option>');
        }   
        
        // Build month select 
        this.$Objs.monthSelect = SlimWidgets.Util.create$Obj(this.options.container.dateSelectsWrapper.monthSelect);
        
        for (var m = 0; m < this.options.monthNames.length; m++){
            this.$Objs.monthSelect.append('<option value="' + m + '">' + this.options.monthNames[m] + '</option>');
        } 
        
        // Build day select
        this.$Objs.daySelect = SlimWidgets.Util.create$Obj(this.options.container.dateSelectsWrapper.daySelect);
         
        
        // appending a dot after day in the German Version
        if (this.options.lang === 'de') {
            
            for (var d = 1; d <= 31; d++){
                this.$Objs.daySelect.append('<option value="' + d + '">' + ('\xa0'  + d).slice(-2) + '.</option>');
            }
        }
        
        else {
            
            for (var d = 1; d <= 31; d++){ this.$Objs.daySelect.append('<option value="' + d + '">' + ('\xa0'  + d).slice(-2) + '</option>')}
        }
            
        
        // Append selects
        if (this.options.format.match(/DD.MM.YYYY/)){
        
            this.$Objs.dateSelectsWrapper.append([this.$Objs.daySelect, this.$Objs.monthSelect, this.$Objs.yearSelect]);
            
        } else if (this.options.format.match(/MM.DD.YYYY/)) {
        
             this.$Objs.dateSelectsWrapper.append([this.$Objs.monthSelect, this.$Objs.daySelect, this.$Objs.yearSelect]);
            
        } else {
             this.$Objs.dateSelectsWrapper.append([this.$Objs.yearSelect, this.$Objs.monthSelect, this.$Objs.daySelect]);
        }
        
        // Confirm button
        this.$Objs.confirmBtn = SlimWidgets.Util.create$Obj(this.options.container.confirmBtn)
            .text(this.options.textConfirmBtn)
            .appendTo(this.$Objs.container);
        
        // Cancel button
        this.$Objs.cancelBtn = SlimWidgets.Util.create$Obj(this.options.container.cancelBtn)
            .text(this.options.textCancelBtn)
            .appendTo(this.$Objs.container);  
    };

    // Inherit prototype from SlimWidgets.Overlay
    SlimWidgets.Datepicker.prototype = SlimWidgets.Util.createObject(SlimWidgets.Overlay.prototype);
    SlimWidgets.Datepicker.prototype.constructor = SlimWidgets.Datepicker;  
    
    // Activate buttons. All handlers defined here will be removed when function Datepicker.remove is called 
    // (by jQuery.remove). They will be re-appended when function Datepicker.append is called (which re-envokes _hooks)
    SlimWidgets.Datepicker.prototype._hooks = function(){
        
        this.removeOnClickTo(this.$Objs.closeBtn);

        // element input already set
        if (this.$Objs.target && this.$Objs.target.length !== 0){
            
             var that = this; 
            
            // Confirm button
            this.$Objs.confirmBtn.on('click', function(){
                
                var date = that.getDate();
                
                if (date){
                    that.$Objs.target.val(date).text(date); 
                    that.remove(); 
                } else if (that.options.invalidDateMsg){
                    alert(that.options.invalidDateMsg)
                }
                // else -> nothing happens when an invalid date is selected
            });
            
            // Cancel button
            this.$Objs.cancelBtn.on('click', function(){
                
                that.$Objs.target.val('');
                that.remove(); 
            }); 
            
            // Update element input on select change
            if (this.options.updateOnChange){
                
                $('.' + this.options.container.dateSelectsWrapper._className + ' select').on('change', function(){
                
                    var date = that.getDate();
                    if (date){ that.$Objs.target.val(date).text(date); }
                })
            };
            
        // No element input
        } else {
            
            this.removeOnClickTo(this.$Objs.confirmBtn);
            this.removeOnClickTo(this.$Objs.cancelBtn);
        }
        
        this.setDate();
    };
    
    // Determine a target that will be controlled by the datepicker
    SlimWidgets.Datepicker.prototype.setTarget = function(selector){
    
      this.$Objs.target = $(selector); 
      return this;
    };  
    
    // Unset target
    SlimWidgets.Datepicker.prototype.unsetTarget = function(){
    
      this.$Objs.target = (function(){}()); 
      return this;
    };  
    
    // Get the currently selected date
    SlimWidgets.Datepicker.prototype.getDate = function(){
        
        var delimeters = this.options.format.replace(/[YMD]/g, ''); 
        var d1 = delimeters.substring(0,1);
        var d2 = delimeters.substring(1,2);
        
        var year = this.$Objs.yearSelect.val(); 
        var month = this.$Objs.monthSelect.val(); 
        var day = this.$Objs.daySelect.val(); 
        
        var date = new Date(year, month, day);
            
        // Check if date is valid
        if (date.getDate() != day || date.getMonth() != month || date.getFullYear() != year) {   
            return false;
        }   
            
        // Pad month and day numbers
        month = ('0' + (parseInt(month) + 1)).slice(-2);
        day = ('0' + day).slice(-2); 
        
        if (this.options.format.match(/DD.MM.YYYY/)){
        
            return day + d1 + month + d2 + year;
        
        } else if (this.options.format.match(/MM.DD.YYYY/)) {
        
            return month + d1 + day + d2 + year;
             
        } else {    // YYYY.MM.DD
        
             return  year + d1 + month + d2 + day;
        }
    }; 
    
    // Set date (from paramter date or from the associated input) to the datepicker date
    SlimWidgets.Datepicker.prototype.setDate = function(date){
        
        var date = date || this.$Objs.target.val() || ''; 
        var dateParts; 
        var year;
        var month;
        var day; 
        
        dateParts = date.split(/[-_\.\\]/);
        
        if (dateParts.length != 3) { return false; }
        
        if (this.options.format.match(/DD.MM.YYYY/)){
        
            day = dateParts[0];
            month = dateParts[1];
            year = dateParts[2];
        
        } else if (this.options.format.match(/MM.DD.YYYY/)) {
        
            day = dateParts[1]
            month = dateParts[0];
            year = dateParts[2]
        
        } else {     // YYYY.MM.DD
            
            day = dateParts[2]
            month = dateParts[1];
            year = dateParts[0]
        }
        
        this.$Objs.daySelect.val(parseInt(day));
        this.$Objs.monthSelect.val(parseInt(month)-1);
        this.$Objs.yearSelect.val(year);
    };
    
    
    /* --------------------------------------------------------------------------------------------------------------
     * TOOLTIP
     * -------------------------------------------------------------------------------------------------------------- */
    
    /* Tooltips providing additional information when hovering their parent element (set via 'setParent'). 
     * They may not only contain plain text (retrieved from the title-attribute of the parent element) 
     * but also custom HTML content. */
    SlimWidgets.Tooltip = function(content, options){

        SlimWidgets.Core.call(this, 'tooltip', options);
        
        this.$Objs.container; 
        this.$Objs.parent; 
        this.$Objs.closeBtn;
        this.appendDelayed; 
        
        if (this.options.withCloseBtn){
        
            this.$Objs.closeBtn = SlimWidgets.Util.create$Obj(this.options.closeBtn);
            this.$Objs.main.append(this.$Objs.closeBtn)
                .addClass(this.options._addClassNames.withCloseBtn);
        }
        
        if (this.options.withArrow){
            this.$Objs.main.addClass(this.options._addClassNames.withArrow);
        }
        
        this.setContent(content);
    };    
    
    // Determine popup position (according to parent if possible)
    SlimWidgets.Tooltip.prototype._getPosition = function(){
        
        var top = this.options.offset.top || 0; 
        var left = this.options.offset.left || 0;  
        
        // Determine position according to parent
        if (this.$Objs.parent && this.$Objs.parent.length !== 0){
            top += this.$Objs.parent.offset().top + this.$Objs.parent.outerHeight();
            left += this.$Objs.parent.offset().left; 
        }
        
        return {top:top, left:left};
    };
    
    /* Set a parent element which provides text content and and position for the tooltip.
     * Any previously set tooltip content will be overridden with the value of parents title attribute 
     * (if existent). */
    SlimWidgets.Tooltip.prototype.setParent = function(selector){
    
        // Second option required in cases when attribute title has been removed in previous calls to function setParent
        var title = $(selector).attr('title') || $(selector).attr('data-slim-title'); 
    
        this.$Objs.parent = $(selector);
        
        if (title){
            this.$Objs.parent.attr('data-slim-title', title);
            this.$Objs.parent.removeAttr('title');
            this.setContent(title); 
        }
        
        return this; 
    };
    
    // Remove a previously set parent element
    SlimWidgets.Tooltip.prototype.unsetParent = function(){
        
        var title = this.$Objs.parent.attr('data-slim-title');
        
        if (title){
            this.$Objs.parent.attr('title', title);
            this.$Objs.parent.removeAttr('data-slim-title');  
        }
        
        this.$Objs.parent = (function(){}());
        
        return this; 
    };
    
    // Set tooltip content
    SlimWidgets.Tooltip.prototype.setContent = function(content){ 

        if (!content || content.length === 0){ return this; }
        
        // Insert content into existing container or ...
        if (this.$Objs.container){
        
            this.$Objs.container.empty().append(content); 
            
        } else {    // ... build new container and append content
        
            this.$Objs.container = SlimWidgets.Util.create$Obj(this.options.container).append(content);
            this.$Objs.main.append(this.$Objs.container); 
        }
        
        return this;
    };
    
    // Remove tooltip content
    SlimWidgets.Tooltip.prototype.removeContent = function(){ 
    
        if (this.$Objs.container){
            this.$Objs.container.remove(); 
            this.$Objs.container = (function(){}());
        }
    
        return this;
    };
    
    // Actually append the tooltip
    SlimWidgets.Tooltip.prototype._append = function(){ 
    
        this.$Objs.main.css(this._getPosition()).show().appendTo('body'); 
    
        if (this.options.hideOnClick){
        
            var that = this;
        
            $(window).on('click.' + that.slimId, function(e){ 
                // If not clicked (children of) tooltip
                if ($(e.target).closest('.' + that.options._className).get(0) !== that.$Objs.main.get(0)){ that.remove(); }
            });
        }
        
        if (this.$Objs.closeBtn){
        
            var that = this; 
        
            this.$Objs.closeBtn.on('click', function(e){ 
                e.stopPropagation(); 
                that.remove();
            });
        }
          
        return this;
    }; 
    
    // Append the tooltip directly or delayed
    SlimWidgets.Tooltip.prototype.append = function(){
    
        if (this.options.delay) {
        
            var that = this; 
            
            window.clearTimeout(this.appendDelayed);
            this.appendDelayed = window.setTimeout(function() { that._append(); }, that.options.delay);
            
        } else { this._append(); }
    
        return this;
    };
    
    // Remove tooltip
    SlimWidgets.Tooltip.prototype.remove = function(){
        
        if (this.options.delay) {
            window.clearTimeout(this.appendDelayed);
        } 
        
        $(window).off('click.' + this.slimId); 
        this.$Objs.main.remove();
        
    };
    
    
    /* --------------------------------------------------------------------------------------------------------------
     * PUBLIC ACCESSIBLE FUNCTIONS
     * -------------------------------------------------------------------------------------------------------------- */
    
    window.Slim = {
    
         /* ---- FACTORIES ---- */
        
        BackToTopBtn: function(options){ 
            return new SlimWidgets.BackToTopBtn(options); 
        },
        
        StickyHeader: function(header, options){
            return new SlimWidgets.StickyHeader(header, options);
        },       
        
        Overlay: function(options){
             return new SlimWidgets.Overlay(options); 
        },
        LoadingSpinner: function(options){
             return new SlimWidgets.LoadingSpinner(options); 
        },
        
        Popup: function(contentElements, options, initHooks){
             return new SlimWidgets.Popup(contentElements, options, initHooks); 
        },
        
        Datepicker: function(title, options){
        
            if (options && options.lang && options.lang === 'de'){
                
                return new SlimWidgets.Datepicker(title, $.extend(true, {}, SlimWidgets.LanguageOptions.datepickerGerman, options)); 
            } 
            
            else { return new SlimWidgets.Datepicker(title, $.extend(true, {}, options)); 
                
            }
        
            return new SlimWidgets.Datepicker(title, options); 
        },
        
        Tooltip: function(content, options){
        
            if (!options){  
                // Make it possible to omit parameter 'content'
                if (typeof content === 'object' && !(content instanceof jQuery)){
                    return new SlimWidgets.Tooltip(null, content); 
                }
            }
        
            return new SlimWidgets.Tooltip(content, options); 
        },    
        
        /* ---- FURTHER FUNCTIONS ---- */
        
        /* Bind a popup to one or more selected DOM elements so that the popup is shown
         * as soon as the user clicks one of the elements.*/
        bindPopupTo: function(selector, content, options, initHooks){
            
            $(selector).each(function(){
            
                var $element = $(this);
                var popup; 
                
                $element.on('click.popup', function(){
                
                    if (popup){
                        popup.append(); 
                    } else {
                        popup = new Slim.Popup(content, options, initHooks).append();
                        $element.data('popup', popup);  // // Register widget to be able to unbind it later on
                    } 
                });
            });
        },
        
        /* Like 'Slim.bindPopupTo' except that the popup content is expected to be returned by an user defined ajax-request.
         * A custom callback function (which has to return a Slim.Popup instance) might be defined to handle the ajax response
         * individually. If no callback is defined, the ajax response will directly be displayed as popup.  */
        bindAjaxPopupTo: function(selector, ajax, popupOptions, initHooks, reload){
            
            $(selector).each(function(){
            
                var $element = $(this); 
                var ajaxOptions = $.extend(true, {}, {data: $element.data()}, ajax.options || {}); 
                var loadingSpinner; 
                var popup;      
                
                $element.on('click.popup', function(){
                    
                    if(popup && reload !== true){
                        popup.append(); 
                    } else {
                    
                        if (popup){ popup.remove(); } // Possible if reload == true
                        
                        loadingSpinner = Slim.LoadingSpinner().append();
                        
                        $.ajax(ajaxOptions).done(function(data){ 

                            if (ajax.callback){
                                popup = ajax.callback(data, popupOptions, initHooks); 
                            } else {
                                popup = Slim.Popup({content: data}, popupOptions, initHooks); 
                            };
                            
                            if (popup){
                                $element.data('popup', popup);  // Register widget to be able to unbind it later on
                                popup.append();
                            }
                            
                            loadingSpinner.remove();
                            
                        }).fail(function(){ 
                            loadingSpinner.remove(); 
                        }); 
                    }
                });
            });
        },
        
        // Unbind (and remove) previously bound popups. Works for ajax-popups too!
        unbindPopupFrom: function(selector){
            SlimWidgets.Util.unbindWidgetFrom(selector, 'click.popup', 'popup');
        },
        
        /* Bind a datepicker to one or more selected DOM elements so that the datepicker is shown
         * as soon as one of the elements gets focus. The selected elements will retrieve their value 
         * or text property) directly from the assigned datepicker. */
        bindDatepickerTo: function(selector, title, datepickerOptions, initDate){
            
            $(selector).each(function(){
                
                var $element = $(this); 
                var datepicker;
                
                $element.on('focus.datepicker', function(){
                    
                    if (datepicker){
                        datepicker.append();
                    } else {
                        datepicker = Slim.Datepicker(title, datepickerOptions).setTarget(this).append(); 
                        $element.data('datepicker', datepicker);    // Register widget to be able to unbind it later on
                        
                        if (initDate){ datepicker.setDate(initDate); }
                    }
                }); 
            });
        },
     
        // Unbind (and remove) previously bound datepickers.
        unbindDatepickerFrom: function(selector){
            SlimWidgets.Util.unbindWidgetFrom(selector, 'focus.datepicker', 'datepicker');
        },
     
        /* Bind a tooltip to one or more selected DOM elements which provide text content and position for the 
         * tooltip. The tooltip is shown as soon as the parent element is hovered and disapears again on mouseout. */
        bindTooltipTo: function(selector, content, tooltipOptions){
        
            // Exclude mobile devices
            if( /Android|webOS|iPhone|iPad|iPod|BlackBerry/i.test(navigator.userAgent) ) {
                return false;  
            }
            
            $(selector).each(function(){
                
                var $element = $(this); 
                var tooltip;
                
                $element.on('mouseover.tooltip', function(e){       
      
                    if (tooltip){   
                        tooltip.append(e);
                    } else {
                        tooltip = Slim.Tooltip(content, tooltipOptions).setParent(this).append(e);
                        $element.data('tooltip', tooltip);  // Register widget to be able to unbind it later on
                    } 
                    
                }).on('mouseout.tooltip', function(){
                    
                    if (tooltip){ 
                        tooltip.remove(); 
                    }
                }); 
            }); 
        }, 
        
        // Unbind (and remove) previously bound tooltips.
        unbindTooltipFrom: function(selector){
        
            $(selector).each(function(){
                
                if ($(this).data('tooltip')){
                    $(this).data('tooltip').unsetParent();
                }
            });
        
            SlimWidgets.Util.unbindWidgetFrom(selector, 'mouseover.tooltip', 'tooltip');
            SlimWidgets.Util.unbindWidgetFrom(selector, 'mouseout.tooltip', 'tooltip');
        },
        
        bindStaticTooltipTo: function(selector, content, tooltipOptions){
            
            $(selector).each(function(){
                
                var $element = $(this); 
                var staticDefaultOptions = {
                    withCloseBtn:true, 
                    hideOnClick:false,
                    delay:false
                };
                var options; 
                var tooltip;
                
                if (!tooltipOptions){  
                    // Make it possible to omit parameter 'content'
                    if (typeof content === 'object' && !(content instanceof jQuery)){
                        options = $.extend(true, {}, staticDefaultOptions, content);
                    } else {
                        options = staticDefaultOptions;
                    }
                } else {
                    options = $.extend(true, {}, staticDefaultOptions, tooltipOptions);
                }
                
                // Create on mouseover
                $element.one('mouseover.tooltip', function(){
                    tooltip = Slim.Tooltip(content, options).setParent(this); 
                    $element.data('tooltip', tooltip);  // Register widget to be able to unbind it later on
                })
                
                // Show/hide on click
                $element.on('click.tooltip', function(e){
                    
                    e.stopPropagation(); // -> especially important if 'hideOnClick'=true
                    
                    // if not already visible -> hide
                    if (!tooltip.$Objs.main.is(":visible")){
                        tooltip.append();
                    } else {
                        tooltip.remove();
                    }
                }) 
            })
        },
        
        // Unbind (and remove) previously bound static tooltips.
        unbindStaticTooltipFrom: function(selector){
            SlimWidgets.Util.unbindWidgetFrom(selector, 'mouseover.tooltip', 'tooltip');
            SlimWidgets.Util.unbindWidgetFrom(selector, 'click.tooltip', 'tooltip');
        },
         
        // Create an autocomplete input 
        createAutocomplteInput: function(inputSelector, options){ 

            var defaultOptions = {
                delay: 500,
                minInput: 1,
                container: {
                    className: 'autocomplete-container',
                    delay: false,
                    withCloseBtn: true, 
                    withArrow: true,
                    offset: {
                        left: 0,
                        top:10
                    },
                    adaptToInput: true
                },
                ajaxUrl: ''     // set via options
            };
            var options = $.extend(true, {}, defaultOptions, options); 
            var $input = $(inputSelector);
            var container = Slim.Tooltip(options.container).setParent($input);
            var containerWidth; 
            var timer;
            
            $input.attr('autocomplete', 'off');
            
            // Adapt container width to input width
            if (options.container.adaptToInput){
                
                containerWidth = $input.outerWidth() || 400;
                container.$Objs.main.css({
                    maxWidth: containerWidth,
                    width: containerWidth
                });
            }
            
            $input.on('click input', function(e){

                e.stopPropagation(); 
                window.clearTimeout(timer);
            
                var $input = $(this); 
                var delay = e.type === 'click' ? 0 : options.delay;
        
                // Prevent autocomplete requests with less than x letters
                if ($input.val().length <= options.minInput || 0){   
                
                    container.removeContent().remove(); 
                 
                } else {
                 
                    //  Send delayed ajax request
                    timer = window.setTimeout(function(){   
                
                        window.clearTimeout(timer);
                        
                        var data = {};
                        data[$input.attr('id') || $input.attr('name')] = $input.val();  

                        $.ajax({
                            url: options.ajaxUrl,
                            cache: false,
                            data: data
                        }).done(function(data){
                            
                            if (data){
                                container.setContent(data).append();
                            } 
                        });
                    
                    }, delay);
                }
            });  
        }
    }
}
(jQuery));